@@ -61,6 +61,7 @@ const char *flatpak_context_sockets[] = {
   "fallback-x11",
   "ssh-auth",
   "pcsc",
+  "cups",
   NULL
 };
 
@@ -68,6 +69,7 @@ const char *flatpak_context_devices[] = {
   "dri",
   "all",
   "kvm",
+  "shm",
   NULL
 };
 
@@ -1242,7 +1244,7 @@ static GOptionEntry context_options[] = {
   { "system-no-talk-name", 0, G_OPTION_FLAG_IN_MAIN, G_OPTION_ARG_CALLBACK, &option_system_no_talk_name_cb, N_("Don't allow app to talk to name on the system bus"), N_("DBUS_NAME") },
   { "add-policy", 0, G_OPTION_FLAG_IN_MAIN, G_OPTION_ARG_CALLBACK, &option_add_generic_policy_cb, N_("Add generic policy option"), N_("SUBSYSTEM.KEY=VALUE") },
   { "remove-policy", 0, G_OPTION_FLAG_IN_MAIN, G_OPTION_ARG_CALLBACK, &option_remove_generic_policy_cb, N_("Remove generic policy option"), N_("SUBSYSTEM.KEY=VALUE") },
-  { "persist", 0, G_OPTION_FLAG_IN_MAIN, G_OPTION_ARG_CALLBACK, &option_persist_cb, N_("Persist home directory"), N_("FILENAME") },
+  { "persist", 0, G_OPTION_FLAG_IN_MAIN, G_OPTION_ARG_CALLBACK, &option_persist_cb, N_("Persist home directory subpath"), N_("FILENAME") },
   /* This is not needed/used anymore, so hidden, but we accept it for backwards compat */
   { "no-desktop", 0, G_OPTION_FLAG_IN_MAIN |  G_OPTION_FLAG_HIDDEN, G_OPTION_ARG_NONE, &option_no_desktop_deprecated, N_("Don't require a running session (no cgroups creation)"), NULL },
   { NULL }
@@ -1767,6 +1769,128 @@ flatpak_context_get_needs_system_bus_proxy (FlatpakContext *context)
   return g_hash_table_size (context->system_bus_policy) > 0;
 }
 
+static gboolean
+adds_flags (guint32 old_flags, guint32 new_flags)
+{
+  return (new_flags & ~old_flags) != 0;
+}
+
+static gboolean
+adds_bus_policy (GHashTable *old, GHashTable *new)
+{
+  GLNX_HASH_TABLE_FOREACH_KV (new, const char *, name, gpointer, _new_policy)
+    {
+      int new_policy = GPOINTER_TO_INT (_new_policy);
+      int old_policy = GPOINTER_TO_INT (g_hash_table_lookup (old, name));
+      if (new_policy > old_policy)
+        return TRUE;
+    }
+
+  return FALSE;
+}
+
+static gboolean
+adds_generic_policy (GHashTable *old, GHashTable *new)
+{
+  GLNX_HASH_TABLE_FOREACH_KV (new, const char *, key, GPtrArray *, new_values)
+    {
+      GPtrArray *old_values = g_hash_table_lookup (old, key);
+      int i;
+
+      if (new_values == NULL || new_values->len == 0)
+        continue;
+
+      if (old_values == NULL || old_values->len == 0)
+        return TRUE;
+
+      for (i = 0; i < new_values->len; i++)
+        {
+          const char *new_value = g_ptr_array_index (new_values, i);
+
+          if (!flatpak_g_ptr_array_contains_string (old_values, new_value))
+            return TRUE;
+        }
+    }
+
+  return FALSE;
+}
+
+static gboolean
+adds_filesystem_access (GHashTable *old, GHashTable *new)
+{
+  FlatpakFilesystemMode old_host_mode = GPOINTER_TO_INT (g_hash_table_lookup (old, "host"));
+
+  GLNX_HASH_TABLE_FOREACH_KV (new, const char *, location, gpointer, _new_mode)
+    {
+      FlatpakFilesystemMode new_mode = GPOINTER_TO_INT (_new_mode);
+      FlatpakFilesystemMode old_mode = GPOINTER_TO_INT (g_hash_table_lookup (old, location));
+
+      /* Allow more limited access to the same thing */
+      if (new_mode <= old_mode)
+        continue;
+
+      /* Allow more limited access if we used to have access to everything */
+     if (new_mode <= old_host_mode)
+        continue;
+
+     /* For the remainder we have to be pessimistic, for instance even
+        if we have home access we can't allow adding access to ~/foo,
+        because foo might be a symlink outside home which didn't work
+        before but would work with an explicit access to that
+        particular file. */
+
+      return TRUE;
+    }
+
+  return FALSE;
+}
+
+
+gboolean
+flatpak_context_adds_permissions (FlatpakContext *old,
+                                  FlatpakContext *new)
+{
+  guint32 old_sockets;
+
+  if (adds_flags (old->shares & old->shares_valid,
+                  new->shares & new->shares_valid))
+    return TRUE;
+
+  old_sockets = old->sockets & old->sockets_valid;
+
+  /* If we used to allow X11, also allow new fallback X11,
+     as that is actually less permissions */
+  if (old_sockets & FLATPAK_CONTEXT_SOCKET_X11)
+    old_sockets |= FLATPAK_CONTEXT_SOCKET_FALLBACK_X11;
+
+  if (adds_flags (old_sockets,
+                  new->sockets & new->sockets_valid))
+    return TRUE;
+
+  if (adds_flags (old->devices & old->devices_valid,
+                  new->devices & new->devices_valid))
+    return TRUE;
+
+  /* We allow upgrade to multiarch, that is really not a huge problem */
+  if (adds_flags ((old->features & old->features_valid) | FLATPAK_CONTEXT_FEATURE_MULTIARCH,
+                  new->features & new->features_valid))
+    return TRUE;
+
+  if (adds_bus_policy (old->session_bus_policy, new->session_bus_policy))
+    return TRUE;
+
+  if (adds_bus_policy (old->system_bus_policy, new->system_bus_policy))
+    return TRUE;
+
+  if (adds_generic_policy (old->generic_policy, new->generic_policy))
+    return TRUE;
+
+  if (adds_filesystem_access (old->filesystems, new->filesystems))
+    return TRUE;
+
+  return FALSE;
+}
+
 gboolean
 flatpak_context_allows_features (FlatpakContext        *context,
                                  FlatpakContextFeatures features)
@@ -1831,6 +1955,7 @@ void
 flatpak_context_add_bus_filters (FlatpakContext *context,
                                  const char     *app_id,
                                  gboolean        session_bus,
+                                 gboolean        sandboxed,
                                  FlatpakBwrap   *bwrap)
 {
   GHashTable *ht;
@@ -1840,8 +1965,13 @@ flatpak_context_add_bus_filters (FlatpakContext *context,
   flatpak_bwrap_add_arg (bwrap, "--filter");
   if (app_id && session_bus)
     {
-      flatpak_bwrap_add_arg_printf (bwrap, "--own=%s.*", app_id);
-      flatpak_bwrap_add_arg_printf (bwrap, "--own=org.mpris.MediaPlayer2.%s.*", app_id);
+      if (!sandboxed)
+        {
+          flatpak_bwrap_add_arg_printf (bwrap, "--own=%s.*", app_id);
+          flatpak_bwrap_add_arg_printf (bwrap, "--own=org.mpris.MediaPlayer2.%s.*", app_id);
+        }
+      else
+        flatpak_bwrap_add_arg_printf (bwrap, "--own=%s.Sandboxed.*", app_id);
     }
 
   if (session_bus)
